SystemOrganization addCategory: #Gofer!

Object subclass: #Gofer
	instanceVariableNames: 'packages repository'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Gofer'!

!Gofer commentStamp: 'lr 8/17/2009 13:36' prior: 0!
Gofer loads, merges and unloads Monticello packages.

When loading a package with Gofer, it ...
- cleanly loads one or more packages from one or more repositories,
- finds the "latest" version of a package matching the given name prefix,
- makes sure that dependencies are correctly loaded as well,
- makes sure that the repository is assigned to the working copy, and
- makes sure that the repository is shared among all packages of the same repository. 

When unloading a package with Gofer, it ...
- cleanly unloads one or more packages from the image,
- detects dependencies and automatically unloads these as well,
- removes class categories and method protocols that belonged to the packages,
- unregisters repositories that are no longer in use.

" Example: Load Seaside 2.8 "
Gofer new
	url: 'http://www.squeaksource.com/KomHttpServer';
	package: 'DynamicBindings';
	package: 'KomServices';
	package: 'KomHttpServer';
	url: 'http://www.squeaksource.com/Seaside';
	package: 'Seaside2.8a';
	package: 'Scriptaculous';
	load.
	
" Example: Unload Seaside 2.8 "
Gofer new
	package: 'DynamicBindings';
	package: 'KomServices';
	package: 'KomHttpServer';
	package: 'Seaside2';
	package: 'Scriptaculous';
	unload.!

!Gofer methodsFor: 'options-repositories' stamp: 'lr 7/10/2009 16:27'!
croquet: aString
	self url: 'http://hedgehog.software.umn.edu:8888/' , aString! !

!Gofer methodsFor: 'options-repositories' stamp: 'lr 7/10/2009 16:27'!
impara: aString
	self url: 'http://source.impara.de/' , aString! !

!Gofer methodsFor: 'initialization' stamp: 'lr 8/17/2009 13:46'!
initialize
	packages := OrderedCollection new! !

!Gofer methodsFor: 'actions' stamp: 'lr 8/17/2009 14:39'!
load
	"Load the specified packages."
	
	^ (GoferLoad on: self packages) execute! !

!Gofer methodsFor: 'actions' stamp: 'lr 8/17/2009 14:39'!
merge
	"Merge the specified packages."
	
	^ (GoferMerge on: self packages) execute! !

!Gofer methodsFor: 'adding' stamp: 'lr 8/19/2009 10:27'!
package: aString
	self packages addLast: (GoferPackage name: aString repository: self repository)! !

!Gofer methodsFor: 'accessing' stamp: 'lr 8/17/2009 13:46'!
packages
	^ packages! !

!Gofer methodsFor: 'options-repositories' stamp: 'lr 7/10/2009 16:25'!
renggli: aString
	self url: 'http://source.lukas-renggli.ch/' , aString! !

!Gofer methodsFor: 'accessing' stamp: 'lr 7/4/2009 18:06'!
repository
	^ repository! !

!Gofer methodsFor: 'options' stamp: 'lr 7/5/2009 10:20'!
repository: aRepository
	"Set the reposiory aRepository to load from during the following load actions."

	MCRepositoryGroup default
		addRepository: aRepository.
	repository := MCRepositoryGroup default repositories
		detect: [ :each | each = aRepository ]
		ifNone: [ nil ]! !

!Gofer methodsFor: 'options-repositories' stamp: 'lr 7/10/2009 16:29'!
saltypickle: aString
	self url: 'http://squeak.saltypickle.com/' , aString! !

!Gofer methodsFor: 'options-repositories' stamp: 'lr 7/10/2009 16:28'!
squeakfoundation: aString
	self url: 'http://source.squeakfoundation.org/' , aString! !

!Gofer methodsFor: 'options-repositories' stamp: 'lr 7/10/2009 16:28'!
squeaksource: aString
	self url: 'http://www.squeaksource.com/' , aString! !

!Gofer methodsFor: 'actions' stamp: 'lr 8/19/2009 11:50'!
unload
	"Unload the specified packages."
	
	^ (GoferUnload on: self packages) execute! !

!Gofer methodsFor: 'options' stamp: 'lr 7/5/2009 09:58'!
url: anUrlString
	"Convenience method to set a repository URL using anUrlString."

	self repository: (MCHttpRepository
		location: anUrlString
		user: String new
		password: String new)! !

!Gofer methodsFor: 'options-repositories' stamp: 'lr 7/10/2009 16:26'!
wiresong: aString
	self url: 'http://source.wiresong.ca/' , aString! !

Object subclass: #GoferOperation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Gofer'!

GoferOperation subclass: #GoferLoad
	instanceVariableNames: 'loader versions repositories'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Gofer'!

!GoferLoad methodsFor: 'private' stamp: 'lr 8/19/2009 12:15'!
addPackage: aPackage
	| version |
	version := aPackage findLatestVersion.
	version withAllDependenciesDo: [ :dependency | repositories at: dependency put: aPackage repository ].
	self loader addVersion: version! !

!GoferLoad methodsFor: 'private' stamp: 'lr 8/19/2009 12:17'!
attachRepositories
	repositories keysAndValuesDo: [ :version :repository |
		version workingCopy repositoryGroup
			addRepository: repository ]! !

!GoferLoad methodsFor: 'running' stamp: 'lr 8/19/2009 12:18'!
execute
	self loadPackages.
	self attachRepositories! !

!GoferLoad methodsFor: 'initialization' stamp: 'lr 8/19/2009 12:16'!
initialize
	super initialize.
	repositories := Dictionary new! !

!GoferLoad methodsFor: 'initialization' stamp: 'lr 8/19/2009 12:18'!
initializeOn: aCollection
	super initializeOn: aCollection.
	aCollection do: [ :package | self addPackage: package ]! !

!GoferLoad methodsFor: 'private' stamp: 'lr 8/19/2009 12:18'!
loadPackages
	self loader load! !

!GoferLoad methodsFor: 'accessing' stamp: 'lr 8/19/2009 12:10'!
loader
	"Answer the monticello loader object."
	
	^ loader ifNil: [ loader := MCVersionLoader new ]! !

GoferLoad subclass: #GoferMerge
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Gofer'!

!GoferMerge methodsFor: 'private' stamp: 'lr 8/19/2009 12:18'!
loadPackages
	"Merge the versions and avoid a dialog if there are no conflicts."

	[ self loader merge ]
		on: MCMergeResolutionRequest
		do: [ :request |
			request merger conflicts isEmpty
				ifTrue: [ request resume: true ]
				ifFalse: [ request pass ] ]! !

!GoferMerge methodsFor: 'accessing' stamp: 'lr 8/19/2009 12:11'!
loader
	"Answer the monticello loader object."
	
	^ loader ifNil: [ loader := MCVersionMerger new ]! !

!GoferOperation class methodsFor: 'instance creation' stamp: 'lr 8/19/2009 11:50'!
on: aCollection
	^ self basicNew initializeOn: aCollection! !

!GoferOperation methodsFor: 'private' stamp: 'lr 8/19/2009 11:23'!
cleanup
	"Some generic cleanup code."

	MCDefinition clearInstances.
	MCFileBasedRepository flushAllCaches.
	MCWorkingCopy changed: #allManagers! !

!GoferOperation methodsFor: 'running' stamp: 'lr 8/17/2009 14:40'!
execute
	"Execute the receiving action."

	self subclassResponsibility! !

!GoferOperation methodsFor: 'initialization' stamp: 'lr 8/19/2009 11:02'!
initialize! !

!GoferOperation methodsFor: 'initialization' stamp: 'lr 8/19/2009 11:50'!
initializeOn: aCollection
	self initialize! !

GoferOperation subclass: #GoferUnload
	instanceVariableNames: 'unloader workingCopies'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Gofer'!

!GoferUnload methodsFor: 'private' stamp: 'lr 8/19/2009 11:41'!
addWorkingCopy: aWorkingCopy
	(workingCopies includes: aWorkingCopy)
		ifTrue: [ ^ self ].
	workingCopies addLast: aWorkingCopy.
	aWorkingCopy requiredPackages
		reverseDo: [ :each | self addWorkingCopy: each workingCopy ]! !

!GoferUnload methodsFor: 'private' stamp: 'lr 8/19/2009 11:21'!
cleanupCategories: aWorkginCopy
	"Cleanup class categories, leftover from the unloading process."

	aWorkginCopy packageInfo systemCategories do: [ :category |
		(SystemOrganization classesInCategory: category) isEmpty
			ifTrue: [ SystemOrganization removeSystemCategory: category ] ]! !

!GoferUnload methodsFor: 'private' stamp: 'lr 8/19/2009 10:58'!
cleanupProtocols: aWorkginCopy
	"Cleanup the method protocols, left over from method extensions."

	aWorkginCopy packageInfo foreignClasses do: [ :class |
		(aWorkginCopy packageInfo foreignExtensionCategoriesForClass: class) do: [ :category |
			(class organization listAtCategoryNamed: category) isEmpty
				ifTrue: [ class organization removeCategory: category ] ] ]! !

!GoferUnload methodsFor: 'running' stamp: 'lr 8/19/2009 11:27'!
execute
	self workingCopies do: [ :copy |
		self unloadPackage: copy.
		self finalizeClasses: copy ].
	self unloader load.
	self workingCopies do: [ :copy |
		self cleanupCategories: copy.
		self cleanupProtocols: copy.
		self unregisterWorkingCopy: copy.
		self unregisterPackageInfo: copy.
		self unregisterRepositories: copy ].
	self cleanup! !

!GoferUnload methodsFor: 'private' stamp: 'lr 8/19/2009 11:05'!
finalizeClasses: aWorkingCopy
	"Call the unload metod of classes if present."

	aWorkingCopy packageInfo classes do: [ :class |
		(class selectors includes: #unload)
			ifTrue: [ class unload ] ]! !

!GoferUnload methodsFor: 'initialization' stamp: 'lr 8/19/2009 11:18'!
initialize
	super initialize.
	unloader := MCMultiPackageLoader new.
	workingCopies := OrderedCollection new! !

!GoferUnload methodsFor: 'initialization' stamp: 'lr 8/19/2009 11:51'!
initializeOn: aCollection
	super initializeOn: aCollection.
	aCollection do: [ :package | self addWorkingCopy: package findWorkingCopy ]! !

!GoferUnload methodsFor: 'private' stamp: 'lr 8/19/2009 11:18'!
unloadPackage: aWorkingCopy
	self unloader unloadPackage: aWorkingCopy package! !

!GoferUnload methodsFor: 'accessing' stamp: 'lr 8/19/2009 11:18'!
unloader
	"Answer the monticello loader object."
	
	^ unloader! !

!GoferUnload methodsFor: 'private' stamp: 'lr 8/19/2009 10:57'!
unregisterPackageInfo: aWorkingCopy
	"Unregister the package information from the system."

	PackageOrganizer default
		unregisterPackage: aWorkingCopy packageInfo! !

!GoferUnload methodsFor: 'private' stamp: 'lr 8/19/2009 10:57'!
unregisterRepositories: aWorkingCopy
	"Remove the repositories if no longer in use by any of the packages."

	aWorkingCopy repositoryGroup repositories allButFirst do: [ :repository |
		MCWorkingCopy allManagers do: [ :copy |
			(copy repositoryGroup includes: repository)
				ifTrue: [ ^ self ] ].
		MCRepositoryGroup default
			removeRepository: repository ]! !

!GoferUnload methodsFor: 'private' stamp: 'lr 8/19/2009 10:58'!
unregisterWorkingCopy: aWorkingCopy
	"Unregister the working copy."

	MCWorkingCopy registry
		removeKey: aWorkingCopy package
		ifAbsent: [ ] ! !

!GoferUnload methodsFor: 'accessing' stamp: 'lr 8/19/2009 11:01'!
workingCopies
	"Answer the working copies to be unloaded."

	^ workingCopies! !

Object subclass: #GoferPackage
	instanceVariableNames: 'name repository'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Gofer'!

!GoferPackage class methodsFor: 'instance creation' stamp: 'lr 8/19/2009 10:27'!
name: aString repository: aRepository
	^ self basicNew initializeName: aString repository: aRepository! !

!GoferPackage methodsFor: 'querying' stamp: 'lr 8/19/2009 11:33'!
findLatestVersion
	| versions |
	versions := self repository allVersionNames
		select: [ :each | each beginsWith: self name ].
	versions := versions asSortedCollection: [ :a :b |
		(a copyAfterLast: $.) asNumber <= (b copyAfterLast: $.) asNumber ].
	versions isEmpty
		ifTrue: [ self error: 'No version named ' , self name printString , ' found' ].
	^ self repository loadVersionFromFileNamed: versions last , '.mcz'! !

!GoferPackage methodsFor: 'querying' stamp: 'lr 8/19/2009 11:32'!
findWorkingCopy
	^ MCWorkingCopy registry
		detect: [ :each | self name beginsWith: each packageName ]
		ifNone: [ self error: 'Working copy ' , self name printString , ' not found' ]! !

!GoferPackage methodsFor: 'initialization' stamp: 'lr 8/19/2009 10:26'!
initializeName: aString repository: aRepository
	name := aString.
	repository := aRepository.
	cache := Dictionary new! !

!GoferPackage methodsFor: 'accessing' stamp: 'lr 8/19/2009 10:30'!
name
	"Answer the package name or prefix."

	^ name! !

!GoferPackage methodsFor: 'printing' stamp: 'lr 8/19/2009 10:30'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' name: '; print: self name! !

!GoferPackage methodsFor: 'accessing' stamp: 'lr 8/17/2009 13:55'!
repository
	"Answer the repository instance of this package, or nil."

	^ repository! !

TestCase subclass: #GoferTest
	instanceVariableNames: 'gofer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Gofer'!

!GoferTest methodsFor: 'running' stamp: 'lr 7/10/2009 16:32'!
setUp
	super setUp.
	gofer := Gofer new! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:51'!
testCroquet
	gofer croquet: 'Hermes'.
	self assert: gofer repository locationWithTrailingSlash = 'http://hedgehog.software.umn.edu:8888/Hermes/'! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:56'!
testImpara
	gofer impara: 'Tweak'.
	self assert: gofer repository locationWithTrailingSlash = 'http://source.impara.de/Tweak/'! !

!GoferTest methodsFor: 'testing' stamp: 'lr 8/19/2009 11:57'!
testLoadUnload
	gofer renggli: 'pieraddons'; package: 'Pier-Setup'.

	self assert: gofer packages size = 1.
	self assert: gofer packages first repository = gofer repository.
	self assert: gofer packages first name = 'Pier-Setup'.

	self shouldnt: [ gofer load ] raise: Error.
	self shouldnt: [ gofer unload ] raise: Error! !

!GoferTest methodsFor: 'testing' stamp: 'lr 8/19/2009 11:57'!
testPackages
	self assert: gofer packages isEmpty.
	
	gofer package: 'Foo'.
	self assert: gofer packages size = 1.
	self assert: gofer packages first name = 'Foo'.
	self assert: gofer packages first repository isNil.
	
	gofer url: 'http://foo.com'; package: 'Bar'.
	self assert: gofer packages size = 2.
	self assert: gofer packages last name = 'Bar'.
	self assert: gofer packages last repository locationWithTrailingSlash = 'http://foo.com/'! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:56'!
testRenggli
	gofer renggli: 'pier'.
	self assert: gofer repository locationWithTrailingSlash = 'http://source.lukas-renggli.ch/pier/'! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:47'!
testRepository
	gofer repository: MCDirectoryRepository new.
	self assert: (gofer repository isKindOf: MCDirectoryRepository)! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:53'!
testSaltypickle
	gofer saltypickle: 'GraphViz'.
	self assert: gofer repository locationWithTrailingSlash = 'http://squeak.saltypickle.com/GraphViz/'! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:57'!
testSqueakfoundation
	gofer squeakfoundation: '39a'.
	self assert: gofer repository locationWithTrailingSlash = 'http://source.squeakfoundation.org/39a/'! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:54'!
testSqueaksource
	gofer squeaksource: 'Seaside29'.
	self assert: gofer repository locationWithTrailingSlash = 'http://www.squeaksource.com/Seaside29/'! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:47'!
testUrl
	gofer url: 'http://source.lukas-renggli.ch/pier'.
	self assert: (gofer repository isKindOf: MCHttpRepository).
	self assert: (gofer repository locationWithTrailingSlash = 'http://source.lukas-renggli.ch/pier/')! !

!GoferTest methodsFor: 'testing-options' stamp: 'lr 7/10/2009 16:55'!
testWiresong
	gofer wiresong: 'ob'.
	self assert: gofer repository locationWithTrailingSlash = 'http://source.wiresong.ca/ob/'! !
